package ethz.ch.pp.mergeSort;

import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveTask;

import ethz.ch.pp.util.ArrayUtils;

public class MergeSortMultiOpt extends RecursiveTask<int[]> {

	private static final long serialVersionUID = 1531647254971804196L;

	private int[] input;
	private int start;
	private int end; // excluded
	private int[] resultFrom;
	private int[] resultTo;

	public MergeSortMultiOpt(int[] input) {
		this.start = 0;
		this.end = input.length;
		this.input = input;
		this.resultFrom = input;
		this.resultTo = new int[end];
	}

	private MergeSortMultiOpt(int start, int end, int[] input, int[] resultFrom, int[] resultTo) {
		this.start = start;
		this.end = end;
		this.input = input;
		this.resultFrom = resultFrom;
		this.resultTo = resultTo;
	}

	public static int[] sort(int[] input, int numThreads) {
		ForkJoinPool fp = new ForkJoinPool(numThreads);
		int[] result = fp.invoke(new MergeSortMultiOpt(input));
		fp.shutdown();
		return result;
	}

	@Override
	protected int[] compute() {
		/*
		 * This is the optimized version of the parallel merge sort that
		 * uses the memory more efficiently. Concretely:
		 *   - The MergeSortMulti creates a new temporary array in each of the recursive steps.
		 *     As a result, the total memory usage is N*log(N)
		 *   - The MergeSortMultiOpt (this class) creates a single additional array of length N 
		 *     at the beginning. Then, all the computation is performed using the original array 
		 *     and the new array, with total memory usage 2*N.
		 * 
		 * Depending on your system, this can lead to ~2x additional speedup.
		 */
		if (end - start <= 2) {
			baseCase();
		} else {
			int mid = start + (end - start) / 2;
			MergeSortMultiOpt t1 = new MergeSortMultiOpt(start, mid, input, resultTo, resultFrom);
			MergeSortMultiOpt t2 = new MergeSortMultiOpt(mid, end, input, resultTo, resultFrom);

			t1.fork();
			t2.compute();
			t1.join();

			ArrayUtils.merge(start, mid, end, resultFrom, resultTo);
		}

		return resultTo;
	}

	private void baseCase() {
		if (end - start == 1) {
			resultTo[start] = input[start];
		} else {
			// assumes that length==2
			if (input[start] > input[start + 1]) {
				// use tmp in case: input == resultTo
				int tmp = input[start];
				resultTo[start] = input[start + 1];
				resultTo[start + 1] = tmp;
			} else {
				resultTo[start] = input[start];
				resultTo[start + 1] = input[start + 1];
			}
		}
	}

}
